@@ -3,12 +3,12 @@
    Network input dispatcher... */
 
 /*
- * Copyright (c) 2004-2010 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004-2022 Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -19,8 +19,8 @@
  * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
  *   Internet Systems Consortium, Inc.
- *   950 Charter Street
- *   Redwood City, CA 94063
+ *   PO Box 360
+ *   Newmarket, NH 03857 USA
  *   <info@isc.org>
  *   https://www.isc.org/
  *
@@ -96,7 +96,7 @@ struct timeval *process_outstanding_timeouts (struct timeval *tvp)
  * to the timer and use that.  The complications arise in trying to ensure
  * that all of the corner cases are covered.  The last one is potentially
  * more painful and requires more investigation.
- * 
+ *
  * The plan is continue with the older DHCP calls and timer list.  The
  * calls will continue to manipulate the list but will also pass a
  * timer to the ISC timer code for the actual dispatch.  Later, if desired,
@@ -110,7 +110,26 @@ dispatch(void)
 {
 	isc_result_t status;
 
-	status = isc_app_ctxrun(dhcp_gbl_ctx.actx);
+	do {
+		status = isc_app_ctxrun(dhcp_gbl_ctx.actx);
+
+		/*
+		 * isc_app_ctxrun can be stopped by receiving a
+		 * signal. It will return ISC_R_RELOAD in that
+		 * case. That is a normal behavior.
+		 */
+
+		if (status == ISC_R_RELOAD) {
+			/*
+			 * dhcp_set_control_state() will do the job.
+			 * Note its first argument is ignored.
+			 */
+			status = dhcp_set_control_state(server_shutdown,
+							server_shutdown);
+			if (status == ISC_R_SUCCESS)
+				status = ISC_R_RELOAD;
+		}
+	} while (status == ISC_R_RELOAD);
 
 	log_fatal ("Dispatch routine failed: %s -- exiting",
 		   isc_result_totext (status));
@@ -174,11 +193,10 @@ isclib_timer_callback(isc_task_t  *taskp,
 
 /* maximum value for usec */
 #define USEC_MAX 1000000
-#define DHCP_SEC_MAX  0xFFFFFFFF
 
 void add_timeout (when, where, what, ref, unref)
 	struct timeval *when;
-	void (*where) PROTO ((void *));
+	void (*where) (void *);
 	void *what;
 	tvref_t ref;
 	tvunref_t unref;
@@ -236,15 +254,19 @@ void add_timeout (when, where, what, ref, unref)
 	 * about the usec value, if it's zero we assume the caller didn't care.
 	 *
 	 * The ISC timer library doesn't seem to like negative values
-	 * and can't accept any values above 4G-1 seconds so we limit
-	 * the values to 0 <= value < 4G-1.  We do it before
-	 * checking the trace option so that both the trace code and
-	 * the working code use the same values.
+	 * and on 64-bit systems, isc_time_nowplusinterval() can generate range
+	 * errors on values sufficiently larger than 0x7FFFFFFF (TIME_MAX), so
+	 * we'll limit the interval to:
+	 *
+	 * 	0 <= interval <= TIME_MAX - 1
+	 *
+	 * We do it before checking the trace option so that both the trace
+	 * code and * the working code use the same values.
 	 */
 
 	sec  = when->tv_sec - cur_tv.tv_sec;
 	usec = when->tv_usec - cur_tv.tv_usec;
-	
+
 	if ((when->tv_usec != 0) && (usec < 0)) {
 		sec--;
 		usec += USEC_MAX;
@@ -253,10 +275,11 @@ void add_timeout (when, where, what, ref, unref)
 	if (sec < 0) {
 		sec  = 0;
 		usec = 0;
-	} else if (sec > DHCP_SEC_MAX) {
-		log_error("Timeout requested too large "
-			  "reducing to 2^^32-1");
-		sec = DHCP_SEC_MAX;
+	} else if (sec >= TIME_MAX) {
+		log_error("Timeout too large "
+			  "reducing to: %lu (TIME_MAX - 1)",
+			  (unsigned long)(TIME_MAX - 1));
+		sec = TIME_MAX - 1;
 		usec = 0;
 	} else if (usec < 0) {
 		usec = 0;
@@ -264,12 +287,12 @@ void add_timeout (when, where, what, ref, unref)
 		usec = USEC_MAX - 1;
 	}
 
-	/* 
+	/*
 	 * This is necessary for the tracing code but we put it
 	 * here in case we want to compare timing information
 	 * for some reason, like debugging.
 	 */
-	q->when.tv_sec  = cur_tv.tv_sec + (sec & DHCP_SEC_MAX);
+	q->when.tv_sec  = cur_tv.tv_sec + sec;
 	q->when.tv_usec = usec;
 
 #if defined (TRACING)
@@ -320,12 +343,12 @@ void add_timeout (when, where, what, ref, unref)
 	q->next  = timeouts;
 	timeouts = q;
 
-	isc_interval_set(&interval, sec & DHCP_SEC_MAX, usec * 1000);
+	isc_interval_set(&interval, sec, usec * 1000);
 	status = isc_time_nowplusinterval(&expires, &interval);
 	if (status != ISC_R_SUCCESS) {
 		/*
-		 * The system time function isn't happy or returned
-		 * a value larger than isc_time_t can hold.
+		 * The system time function isn't happy. Range errors
+		 * should not be possible with the check logic above.
 		 */
 		log_fatal("Unable to set up timer: %s",
 			  isc_result_totext(status));
@@ -352,7 +375,7 @@ void add_timeout (when, where, what, ref, unref)
 }
 
 void cancel_timeout (where, what)
-	void (*where) PROTO ((void *));
+	void (*where) (void *);
 	void *what;
 {
 	struct timeout *t, *q;
