@@ -3,12 +3,12 @@
    Tables of information... */
 
 /*
- * Copyright (c) 2004-2009 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004-2022 Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
  * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
@@ -19,17 +19,11 @@
  * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
  *   Internet Systems Consortium, Inc.
- *   950 Charter Street
- *   Redwood City, CA 94063
+ *   PO Box 360
+ *   Newmarket, NH 03857 USA
  *   <info@isc.org>
  *   https://www.isc.org/
  *
- * This software has been written for Internet Systems Consortium
- * by Ted Lemon in cooperation with Vixie Enterprises and Nominum, Inc.
- * To learn more about Internet Systems Consortium, see
- * ``https://www.isc.org/''.  To learn more about Vixie Enterprises,
- * see ``http://www.vix.com''.   To learn more about Nominum, Inc., see
- * ``http://www.nominum.com''.
  */
 
 #include "dhcpd.h"
@@ -63,9 +57,10 @@ HASH_FUNCTIONS (option_code, const unsigned *, struct option,
        some event.  The special all-ones value means 'infinite'.  May either
        be printed as a decimal, eg, "3600", or as this name, eg, "infinite".
    f - flag (true or false)
-   A - array of whatever precedes (e.g., IA means array of IP addresses)
-   a - array of the preceding character (e.g., IIa means two or more IP
-       addresses)
+   A - array of all that precedes (e.g., fIA means array of records of
+       a flag and an IP address)
+   a - array of the preceding character (e.g., fIa means a single flag
+       followed by an array of IP addresses)
    U - name of an option space (universe)
    F - implicit flag - the presence of the option indicates that the
        flag is true.
@@ -89,10 +84,14 @@ HASH_FUNCTIONS (option_code, const unsigned *, struct option,
        the name of the set of enumeration values to parse or emit,
        followed by a '.'.   The width of the data is specified in the
        named enumeration.   Named enumerations are tracked in parse.c.
-   d - Domain name (i.e., FOO or FOO.BAR).
-   D - Domain list (i.e., example.com eng.example.com)
+   d - Domain name (e.g., FOO or FOO.BAR) no quotes,
+       on-wire format is RFC 1035.
+   D - Domain list (e.g., "example.com eng.example.com") quoted,
+       on-wire format is RFC 1035.
    c - When following a 'D' atom, enables compression pointers.
    Z - Zero-length option
+   k - Key name, unquoted string (e.g. mykey.com or some-text or abc123)
+       parsed with parse_host_name().
 */
 
 struct universe dhcp_universe;
@@ -196,16 +195,32 @@ static struct option dhcp_options[] = {
 
 	{ "client-last-transaction-time", "L",  &dhcp_universe,  91, 1 },
 	{ "associated-ip", "Ia",                &dhcp_universe,  92, 1 },
-#if 0
+#if defined(RFC4578_OPTIONS)
 	/* Defined by RFC 4578 */
-	{ "pxe-system-type", "S",		&dhcp_universe,  93, 1 },
+	{ "pxe-system-type", "Sa",		&dhcp_universe,  93, 1 },
 	{ "pxe-interface-id", "BBB",		&dhcp_universe,  94, 1 },
 	{ "pxe-client-id", "BX",		&dhcp_universe,  97, 1 },
 #endif
 	{ "uap-servers", "t",			&dhcp_universe,  98, 1 },
+#if defined(RFC4776_OPTIONS)
+        { "geoconf-civic", "X",                 &dhcp_universe, 99, 1 },
+#endif
+#if defined(RFC4833_OPTIONS)
+	{ "pcode", "t",				&dhcp_universe, 100, 1 },
+	{ "tcode", "t",				&dhcp_universe, 101, 1 },
+#endif
+#if defined(RFC8925_OPTIONS)
+	{ "v6-only-preferred", "L",		&dhcp_universe, 108, 1 },
+#endif
 	{ "netinfo-server-address", "Ia",	&dhcp_universe, 112, 1 },
 	{ "netinfo-server-tag", "t",		&dhcp_universe, 113, 1 },
 	{ "default-url", "t",			&dhcp_universe, 114, 1 },
+#if defined(RFC2563_OPTIONS)
+	{ "auto-config", "B",			&dhcp_universe, 116, 1 },
+#endif
+#if defined(RFC2937_OPTIONS)
+	{ "name-service-search", "Sa",		&dhcp_universe, 117, 1 },
+#endif
 	{ "subnet-selection", "I",		&dhcp_universe, 118, 1 },
 	{ "domain-search", "Dc",		&dhcp_universe, 119, 1 },
 	{ "vivco", "Evendor-class.",		&dhcp_universe, 124, 1 },
@@ -224,20 +239,47 @@ static struct option dhcp_options[] = {
 	{ "pxe-undefined-7", "X",		&dhcp_universe, 134, 1 },
 	{ "pxe-undefined-8", "X",		&dhcp_universe, 135, 1 },
 #endif
-#if 0
-	/* Not defined by RFC yet */
+#if defined(RFC5192_OPTIONS)
+	{"pana-agent", "Ia",			&dhcp_universe, 136, 1 },
+#endif
+#if defined(RFC5223_OPTIONS)
+	{"v4-lost", "d",			&dhcp_universe, 137, 1 },
+#endif
+#if defined(RFC5417_OPTIONS)
+	{"capwap-ac-v4", "Ia",			&dhcp_universe, 138, 1 },
+#endif
+#if defined(RFC6011_OPTIONS)
+	{ "sip-ua-cs-domains", "Dc",		&dhcp_universe, 141, 1 },
+#endif
+#if defined(RFC6153_OPTIONS)
+	{ "ipv4-address-andsf", "IA",		&dhcp_universe, 142, 1 },
+#endif
+#if defined(RFC6731_OPTIONS)
+        { "rdnss-selection", "BIID",		&dhcp_universe, 146, 1 },
+#endif
+#if defined(RFC5859_OPTIONS)
 	{ "tftp-server-address", "Ia",		&dhcp_universe, 150, 1 },
 #endif
-#if 0
-	/* PXELINUX options: defined by RFC 5071 */
+#if defined(RFC7618_OPTIONS)
+	{ "v4-portparams", "BBS",		&dhcp_universe, 159, 1 },
+#endif
+#if defined(RFC7710_OPTIONS)
+	{ "v4-captive-portal", "t",		&dhcp_universe, 160, 1 },
+#endif
+#if defined(RFC5071_OPTIONS)
+#if  0
+	/* Option 208 has been officially deprecated. Do NOT define it */
 	{ "pxelinux-magic", "BBBB",		&dhcp_universe, 208, 1 },
+#endif
 	{ "loader-configfile", "t",		&dhcp_universe, 209, 1 },
 	{ "loader-pathprefix", "t",		&dhcp_universe, 210, 1 },
 	{ "loader-reboottime", "L",		&dhcp_universe, 211, 1 },
 #endif
-#if 0
-	/* Not defined by RFC yet */
-	{ "vss-info", "BX",			&dhcp_universe, 221, 1 },
+#if defined(RFC5969_OPTIONS)
+        { "option-6rd", "BB6Ia",		&dhcp_universe, 212, 1 },
+#endif
+#if defined(RFC5986_OPTIONS)
+	{"v4-access-domain", "d",		&dhcp_universe, 213, 1 },
 #endif
 	{ NULL, NULL, NULL, 0, 0 }
 };
@@ -331,6 +373,11 @@ static struct option dhcpv6_options[] = {
 	{ "relay-msg", "X",			&dhcpv6_universe,  9, 1 },
 
 	/* Option code 10 is curiously unassigned. */
+	/*
+	 * In draft-ietf-dhc-dhcpv6-25 there were two OPTION_CLIENT_MSG and
+	 * OPTION_SERVER_MSG options. They were eventually unified as
+	 * OPTION_RELAY_MSG, hence no option with value of 10.
+	 */
 #if 0
 	/* XXX: missing suitable atoms for the auth option.  We may want
 	 * to 'virtually encapsulate' this option a la the fqdn option
@@ -396,8 +443,7 @@ static struct option dhcpv6_options[] = {
 
 	/* Note that 35 is not assigned. */
 
-	/* Not yet considering for inclusion. */
-#if 0
+#if defined(RFC4776_OPTIONS)
 			/* RFC4776 OPTIONS */
 
 	{ "geoconf-civic", "X",			&dhcpv6_universe, 36, 1 },
@@ -431,18 +477,20 @@ static struct option dhcpv6_options[] = {
 	{ "fqdn", "Efqdn6-if-you-see-me-its-a-bug-bug-bug.",
 						&dhcpv6_universe, 39, 1 },
 
-	/* Not yet considering for inclusion. */
-#if 0
-			/* draft-ietf-dhc-paa-option-05 */
+
+			/* RFC5192 */
+#if defined(RFC5192_OPTIONS)
 	{ "pana-agent", "6A",			&dhcpv6_universe, 40, 1 },
+#endif
 
 			/* RFC4833 OPTIONS */
-
+#if defined(RFC4833_OPTIONS)
 	{ "new-posix-timezone", "t",		&dhcpv6_universe, 41, 1 },
 	{ "new-tzdb-timezone", "t",		&dhcpv6_universe, 42, 1 },
+#endif
 
 			/* RFC4994 OPTIONS */
-
+#if defined(RFC4994_OPTIONS)
 	{ "ero", "SA",				&dhcpv6_universe, 43, 1 },
 #endif
 
@@ -454,13 +502,94 @@ static struct option dhcpv6_options[] = {
 	{ "lq-relay-data", "6X",		&dhcpv6_universe, 47, 1 },
 	{ "lq-client-link", "6A",		&dhcpv6_universe, 48, 1 },
 
+			/* RFC5223 OPTIONS */
+#if defined(RFC5223_OPTIONS)
+	{ "v6-lost", "d",			&dhcpv6_universe, 51, 1 },
+#endif
+
+			/* RFC5417 OPTIONS */
+#if defined(RFC5417_OPTIONS)
+	{ "capwap-ac-v6", "6a",			&dhcpv6_universe, 52, 1 },
+#endif
+
+			/* RFC5460 OPTIONS */
+#if defined(RFC5460_OPTIONS)
+	{ "relay-id", "X",			&dhcpv6_universe, 53, 1 },
+#endif
+
+			/* RFC5986 OPTIONS */
+#if defined(RFC5986_OPTIONS)
+	{ "v6-access-domain", "d",		&dhcpv6_universe, 57, 1 },
+#endif
+
+			/* RFC6011 OPTIONS */
+#if defined(RFC6011_OPTIONS)
+	{ "sip-ua-cs-list", "D",		&dhcpv6_universe, 58, 1 },
+#endif
+
+			/* RFC5970 OPTIONS */
+#if defined(RFC5970_OPTIONS)
+	{ "bootfile-url", "t",			&dhcpv6_universe, 59, 1 },
+	{ "bootfile-param", "X",		&dhcpv6_universe, 60, 1 },
+	{ "client-arch-type", "SA",		&dhcpv6_universe, 61, 1 },
+	{ "nii", "BBB",				&dhcpv6_universe, 62, 1 },
+#endif
+
+			/* RFC6334 OPTIONS */
+#if defined(RFC6334_OPTIONS)
+	{ "aftr-name", "d",			&dhcpv6_universe, 64, 1 },
+#endif
+
+			/* RFC6440 OPTIONS */
+#if defined(RFC6440_OPTIONS)
+	{ "erp-local-domain-name", "d",		&dhcpv6_universe, 65, 1 },
+#endif
+
+			/* RFC6731 OPTIONS */
+#if defined(RFC6731_OPTIONS)
+	{ "rdnss-selection", "6BD",		&dhcpv6_universe, 74, 1 },
+#endif
+
+			/* RFC6939 OPTIONS */
+#if defined(RFC6939_OPTIONS)
+	{ "client-linklayer-addr", "X",		&dhcpv6_universe, 79, 1 },
+#endif
+
+			/* RFC6977 OPTIONS */
+#if defined(RFC6977_OPTIONS)
+	{ "link-address", "6",			&dhcpv6_universe, 80, 1 },
+#endif
+
+			/* RFC7083 OPTIONS */
+#if defined(RFC7083_OPTIONS)
+	{ "solmax-rt", "L",			&dhcpv6_universe, 82, 1 },
+	{ "inf-max-rt", "L",			&dhcpv6_universe, 83, 1 },
+#endif
+
+			/* RFC7341 OPTIONS */
+#if defined(RFC7341_OPTIONS)
+	{ "dhcpv4-msg", "X",			&dhcpv6_universe, 87, 1 },
+	{ "dhcp4-o-dhcp6-server", "6A",		&dhcpv6_universe, 88, 1 },
+#endif
+
+#if defined(RFC7710_OPTIONS)
+	{ "v6-captive-portal", "t",		&dhcpv6_universe, 103, 1 },
+#endif
+
+	{ "relay-source-port", "S",		&dhcpv6_universe, 135, 1 },
+
+#if defined(RFC6153_OPTIONS)
+	{ "ipv6-address-andsf", "6A",		&dhcpv6_universe, 143, 1 },
+#endif
+
 	{ NULL, NULL, NULL, 0, 0 }
 };
 
 struct enumeration_value dhcpv6_duid_type_values[] = {
-	{ "duid-llt",	DUID_LLT }, /* Link-Local Plus Time */
-	{ "duid-en",	DUID_EN },  /* DUID based upon enterprise-ID. */
-	{ "duid-ll",	DUID_LL },  /* DUID from Link Local address only. */
+	{ "duid-llt",	DUID_LLT },  /* Link-Local Plus Time */
+	{ "duid-en",	DUID_EN },   /* DUID based upon enterprise-ID. */
+	{ "duid-ll",	DUID_LL },   /* DUID from Link Local address only. */
+	{ "duid-uuid",	DUID_UUID }, /* DUID based upon UUID */
 	{ NULL, 0 }
 };
 
@@ -482,6 +611,7 @@ struct enumeration_value dhcpv6_status_code_values[] = {
 	{ "MalformedQuery", 8 }, /* Leasequery not valid.		*/
 	{ "NotConfigured", 9 }, /* The target address is not in config.	*/
 	{ "NotAllowed",  10 }, /* Server doesn't allow the leasequery.	*/
+	{ "QueryTerminated", 11 }, /* Leasequery terminated.		*/
 	{ NULL, 0 }
 };
 
@@ -494,6 +624,9 @@ struct enumeration dhcpv6_status_codes = {
 struct enumeration_value lq6_query_type_values[] = {
 	{ "query-by-address", 1 },
 	{ "query-by-clientid", 2 },
+	{ "query-by-relay-id", 3 },
+	{ "query-by-link-address", 4 },
+	{ "query-by-remote-id", 5 },
 	{ NULL, 0 }
 };
 
@@ -519,6 +652,12 @@ struct enumeration_value dhcpv6_message_values[] = {
 	{ "RELAY-REPL", 13 },
 	{ "LEASEQUERY", 14 },
 	{ "LEASEQUERY-REPLY", 15 },
+	{ "LEASEQUERY-DONE", 16 },
+	{ "LEASEQUERY-DATA", 17 },
+	{ "RECONFIGURE-REQUEST", 18 },
+	{ "RECONFIGURE-REPLY", 19 },
+	{ "DHCPV4-QUERY", 20 },
+	{ "DHCPV4-RESPONSE", 21 },
 	{ NULL, 0 }
 };
 
@@ -539,7 +678,13 @@ const char *dhcpv6_type_names[] = {
 	"Relay-forward",
 	"Relay-reply",
 	"Leasequery",
-	"Leasequery-reply"
+	"Leasequery-reply",
+	"Leasequery-done",
+	"Leasequery-data",
+	"Reconfigure-request",
+	"Reconfigure-reply",
+	"Dhcpv4-query",
+	"Dhcpv4-response"
 };
 const int dhcpv6_type_name_max =
 	(sizeof(dhcpv6_type_names) / sizeof(dhcpv6_type_names[0]));
@@ -559,7 +704,9 @@ static struct option vsio_options[] = {
 struct universe isc6_universe;
 static struct option isc6_options[] = {
 	{ "media", "t",				&isc6_universe,     1, 1 },
-	{ "update-assist", "X",			&isc6_universe,	    2, 1 },
+	{ "update-assist", "X",			&isc6_universe,     2, 1 },
+	{ "4o6-interface", "t",			&isc6_universe, 60000, 1 },
+	{ "4o6-source-address", "6",		&isc6_universe, 60001, 1 },
 	{ NULL, NULL, NULL, 0, 0 }
 };
 
@@ -596,7 +743,7 @@ const char *hardware_types [] = {
 	"unknown-29",
 	"unknown-30",
 	"unknown-31",
-	"unknown-32",
+	"infiniband",
 	"unknown-33",
 	"unknown-34",
 	"unknown-35",
@@ -1402,8 +1549,10 @@ void initialize_common_option_spaces()
 			  &vsio_universe, MDL);
 	universe_hash_add(universe_hash, isc6_universe.name, 0,
 			  &isc6_universe, MDL);
-/* This should not be necessary.  Listing here just for consistency.
- *	universe_hash_add(universe_hash, fqdn6_universe.name, 0,
- *			  &fqdn6_universe, MDL);
- */
+	/* previously this wasn't necessary, now that we can send
+	 * v6 encapsulated options it is.
+	 */
+	universe_hash_add(universe_hash, fqdn6_universe.name, 0,
+			  &fqdn6_universe, MDL);
+
 }
