@@ -1,9 +1,9 @@
 /*
- * Copyright (C) 2006-2016 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2006-2017 by Internet Systems Consortium, Inc. ("ISC")
  *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
  * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
@@ -152,6 +152,9 @@ dhcpleasequery(struct packet *packet, int ms_nulltp) {
 	u_int32_t time_rebinding;
 	u_int32_t time_expiry;
 	u_int32_t client_last_transaction_time;
+#if defined(RELAY_PORT)
+	u_int16_t relay_port = 0;
+#endif
 	struct sockaddr_in to;
 	struct in_addr siaddr;
 	struct data_string prl;
@@ -660,12 +663,20 @@ dhcpleasequery(struct packet *packet, int ms_nulltp) {
 #endif
 	memset(to.sin_zero, 0, sizeof(to.sin_zero));
 
+#if defined(RELAY_PORT)
+	relay_port = dhcp_check_relayport(packet);
+#endif
+
 	/* 
 	 * Leasequery packets are be sent to the gateway address.
 	 */
 	to.sin_addr = packet->raw->giaddr;
 	if (packet->raw->giaddr.s_addr != htonl(INADDR_LOOPBACK)) {
+#if defined(RELAY_PORT)
+		to.sin_port = relay_port ? relay_port : local_port;
+#else
 		to.sin_port = local_port;
+#endif
 	} else {
 		to.sin_port = remote_port; /* XXXSK: For debugging. */
 	}
@@ -862,15 +873,9 @@ valid_query_msg(struct lq6_state *lq) {
 
 exit:
 	if (!ret_val) {
-		if (lq->client_id.len > 0) {
-			data_string_forget(&lq->client_id, MDL);
-		}
-		if (lq->server_id.len > 0) {
-			data_string_forget(&lq->server_id, MDL);
-		}
-		if (lq->lq_query.len > 0) {
-			data_string_forget(&lq->lq_query, MDL);
-		}
+		data_string_forget(&lq->client_id, MDL);
+		data_string_forget(&lq->server_id, MDL);
+		data_string_forget(&lq->lq_query, MDL);
 	}
 	return ret_val;
 }
